import json, re, pytz
from datetime import datetime, timedelta
from types import SimpleNamespace
from django.http import JsonResponse
from django.utils import timezone
from rest_framework import generics, status
from rest_framework.permissions import IsAuthenticated
from rest_framework.decorators import api_view, permission_classes
from clusterCore.models import c123_node, c123_host, c123_inventory, c123_database, Task, c123_cluster
from clusterCore.enums import DBType, ClusterRoles, RestoreOptions, Severity, Status, Inventory, Schedule, CopyOptions, TaskState, Node_Size_Allowed
from clusterCore.api.serializer import c123_HostSerializer, c123_DatabaseSerializer, c123_inventory_serializer, c123_ClusterSerializer, ClusterSizing, c123_NodeSerializer, object_to_json, format_size, get_create_node_response
from .serializer import c123_AddMongoNodeSerializer, c123_CreateMongoClusterSerializer, c123_AddMongoShardedClusterSerializer, c123_AddMongoShardedClusterNodeSerializer, c123_CreateMongoShardedClusterPrimary
from clusterCore.tasks import allow_or_403, checkPathExists, get_cron_schedule, check_space, create_ssh_connection, remote_command, \
cleanup_logs, create_mongo_connection, find_replicaset_primary, drop_mongo_node, append_status_cmd, format_password, get_backup_description, \
export_mongo_node, export_mongo_database, get_avail_change_date
from clusterCore.scheduler import schedule_backup, mongodump_oplog_job, MISFIRE_GRACE_TIME_BACKUP
from clusterCore.exceptions import *
from apscheduler.triggers.interval import IntervalTrigger
from dateutil.parser import parse
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CopyMongoDatabase(request):
    db_id = request.data.get("db_id")
    node_id = request.data.get("node_id")
    source_id = request.data.get("source_id")
    current_user = request.user
    db_option = request.data.get("option")
    #copy_option = 'one_time' or 'live'
    copy_option = request.data.get("copy_option")
    copy_name = request.data.get("copy_name")
    db_exists = request.data.get("db_exists")
    api_call = request.data.get("rest_api")
    if api_call:
        node_target = c123_node.objects.filter(name = node_id, type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value).first()
    else:
        node_target = c123_node.objects.filter(pk = node_id, type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value).first()
    if not node_target:
        raise APIException404("target primary id {} not found".format(node_id))
    allow_or_403(node_target)
    if api_call:
         node_source = c123_node.objects.filter(name = source_id, type = DBType.MONGO.value,  cluster_role = ClusterRoles.PRIMARY.value).first()
    else:
         node_source = c123_node.objects.filter(pk = source_id, type = DBType.MONGO.value,  cluster_role = ClusterRoles.PRIMARY.value).first()
    if not node_source:
        raise APIException404("source primary id {} not found".format(source_id))
    allow_or_403(node_source)
    if api_call:
        db_target = c123_database.objects.filter(name = db_id, node = node_source, type = DBType.MONGO.value, ind_deleted = False).first()
    else:
        db_target = c123_database.objects.filter(pk = db_id, type = DBType.MONGO.value, ind_deleted = False).first()
    if not db_target:
        raise APIException410("target database id {} not available".format(db_id))

    db_source = c123_database.objects.get(pk=db_target.pk)
    if(db_option):
        # here we do for existing database
        match db_option:
            case CopyOptions.OVERRIDE.value:
                copy_name = db_target.name
                db_target.ind_started = False
                db_target.node = node_target
                db_target.database_user = node_target.database_user
                db_target.database_pwd = node_target.database_pwd
                db_target.status = Status.REPLICATING.value
                db_target.save()
            case CopyOptions.COPY.value:
                existing_dbs = c123_database.objects.filter(node = node_target, name = copy_name)
                if not existing_dbs.exists():
                        db_target = c123_database.objects.create(node = node_target, name = copy_name,
                        type = db_source.type, size = db_source.size,
                        database_user = node_target.database_user, database_pwd = node_target.database_pwd,
                        created_by = current_user, last_update_by = current_user,
                        severity = Severity.NOT_OK.value, status = Status.REPLICATING.value
                        )   
                else:
                    undeleted = existing_dbs.filter(ind_deleted = False)
                    if undeleted.exists():
                        db_target = undeleted.first()
                    else:
                        db_target = existing_dbs.first()
                        db_target.ind_deleted = False
                        
                    db_target.type = DBType.MONGO.value
                    db_target.status = Status.REPLICATING.value
                    db_target.save()
                
                db_source.status = Status.COPYING.value
                db_source.ind_started = False
                db_source.save()
    else:
        #here we do for new database
        db_source.status = Status.COPYING.value
        db_source.ind_started = False
        db_source.save()
        db_target = c123_database.objects.create(node = node_target, name = db_source.name,
        type = db_source.type, size = db_source.size,
        database_user = node_target.database_user, database_pwd = node_target.database_pwd,
        created_by = current_user, last_update_by = current_user,
        severity = Severity.NOT_OK.value, status = Status.REPLICATING.value
        )
    tmp_data = json.dumps({'db_source': object_to_json(db_source), 'db_target' : object_to_json(db_target), 'node_source': object_to_json(node_source), 'node_target': object_to_json(node_target), 'host_source': object_to_json(node_source.host), 'host_target': object_to_json(node_target.host), 'db_exists': db_exists, 'db_option': db_option, 'copy_option': copy_option, 'created_by': current_user.id})
    task = Task.objects.create(task='copy_mongo_database_task',
                                displayment=f'copy database {db_source.name} into {node_target.name}',
                                target_type=DBType.MONGO.value,
                                target_id=node_target.id,
                                org_id = node_target.org_id,
                                status=TaskState.PENDING.value,
                                schedule=Schedule.NOW.value,
                                parameters=tmp_data,
                                created_by=current_user)
    
    return append_status_cmd(request, task)

class MongoClusterCreateView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_CreateMongoClusterSerializer
    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)
class MongoShardedClusterAddView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_AddMongoShardedClusterSerializer
    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)
class AddMongoNodeView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_AddMongoNodeSerializer
    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)
class AddMongoShardedClusterNodeView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_AddMongoShardedClusterNodeSerializer
    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)

class AddMongoShardedClusterPrimaryView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_CreateMongoShardedClusterPrimary
    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RestoreMongoDatabase(request):
    """restore a mongo db from dump
    the call is made from client AJAX POST that contains the dump id"""
    dump_id = request.data.get('dump_id')
    api_call = request.data.get('rest_api')
    autostart = True if request.data.get('autostart') == 'true' else False
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    if api_call:
        dump = c123_inventory.objects.filter(name=dump_id, type = Inventory.EXPORT.value).first()
    else:
        dump = c123_inventory.objects.filter(pk=dump_id, type = Inventory.EXPORT.value).first()
    if not dump:
        raise APIException404("dump id {} not found".format(dump_id))
    db = dump.database
    node = db.node
    if db.name == "local":
        raise APIException403("database local cant be restored")
    path = dump.ex_params.get('backup_path')
    path_exists = checkPathExists(app_host, path)
    if not path_exists:
        raise APIException404(f"{path}: no such file or directory")
    # serialize objects for the tasks
    primary = c123_node.objects.filter(cluster = node.cluster, type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
        raise APIException404("primary not found")
    allow_or_403(primary)
    db.status = Status.RESTORING.value
    db.save()
    db_name = db.name
    dump_host = c123_HostSerializer(app_host).data
    node_data = object_to_json(primary)
    host_data = c123_HostSerializer(primary.host).data
    created_by = request.user
    dump = object_to_json(dump)
    tmp_data = json.dumps({'db_id': db.id, 'db_name': db_name, 'node_data': node_data, 'host_data': host_data, 'dump': dump, 'backup_host': dump_host, 'autostart': autostart, 'created_by': created_by.id})
    task = Task.objects.create(task='mongorestore_task',
                            displayment='restore mongo database {}'.format(db.name),
                            target_type=DBType.MONGO.value,
                            target_id=db.id,
                            org_id = primary.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DumpMongoDatabase(request):
    """dump for database"""

    data = request.data
    data['created_by'] = request.user.id
    schedule = data.get('schedule')
    api_call = data.get('api_call')
    node_id = data.get('node_id')
    db_id = data.get('db_id')

    if not schedule:
        task = export_mongo_database(data)
        return append_status_cmd(request, task)
    else:
        from django.conf import settings
        client_timezone = pytz.timezone(data.get("client_timezone"))
        schedule_run_time_str = data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = data.get("days_of_week")
        days_of_month = data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
        if api_call:
            db = c123_database.objects.filter(type = DBType.MONGO.value, name=db_id, node__name = node_id, ind_deleted = False).first()
        else:
            db = c123_database.objects.filter(type = DBType.MONGO.value, pk=db_id, ind_deleted = False).first()
        if not db:
            raise APIException410("database id {} not available".format(db_id))
        node = db.node
        job_id =  f"export_{db.name}_{node.name}"
        scheduler = settings.scheduler
        export_job = next((i for i in scheduler.get_jobs() if job_id in i.id), None)
        if export_job:
            export_job.reschedule(trigger=trigger)
        else:
             settings.scheduler.add_job(export_mongo_database, replace_existing=True, trigger=trigger,  kwargs={"data": data}, id=job_id, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        return JsonResponse({"notification": f"export db {db.name} of {node.name} {description} started"}, status = 200)


@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def MongoDetails(request, id):
    node = c123_node.objects.filter(pk = id, type = DBType.MONGO.value).first()
    if not node:
        raise APIException404("node id {} not found".format(id))
    response = {'node': c123_NodeSerializer(node).data}
    auto_backups = c123_inventory.objects.filter(node=node, type = Inventory.BACKUP.value, is_sched=True)
    if auto_backups.count() > 0:
        if auto_backups.last().status:
            response["auto_backup"] = {
                    'status': "Automatic backup", 'class': "green-star", "detail": "Automatic backup configured"}
        else:
            response["auto_backup"] = {
                    'status': "Automatic backup", 'class': "red-green-star", "detail": "Automatic backup is configured, but the last run reported errors"}
                
    else:
        response["auto_backup"]  = {
                'status': "No automatic backup", 'class': "red-star", "detail" : "Automatic backup is not configured"}
    if node.cluster_role == ClusterRoles.PRIMARY.value:
        primary_backups = c123_inventory_serializer(c123_inventory.objects.filter(node_id = node.id, type = Inventory.BACKUP.value), many = True)
        response["backups"] = primary_backups.data
        primary_dumps = c123_inventory_serializer(c123_inventory.objects.filter(node_id = node.id, type = Inventory.EXPORT.value), many = True)
        response["dumps"] = primary_dumps.data
        standbys = c123_NodeSerializer(c123_node.objects.filter(cluster = node.cluster, type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False), many = True)
        response["nodes"] = standbys.data
        arbiter = c123_node.objects.filter(cluster = node.cluster, type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value,  ind_deleted = False).first()
        if arbiter:
            response["arbiter"] = c123_NodeSerializer(arbiter).data
    
        for i, standby in enumerate(response["nodes"]):
            backups = c123_inventory_serializer(c123_inventory.objects.filter(node_id = standby["id"], type = Inventory.BACKUP.value), many = True)
            response["nodes"][i]["backups"] = backups.data
            dumps = c123_inventory_serializer(c123_inventory.objects.filter(node_id = standby["id"], type = Inventory.EXPORT.value), many = True)
            response["nodes"][i]["dumps"] = dumps.data
    else:
        standbys = c123_NodeSerializer(c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = node.cluster).exclude(pk = node.pk), many = True)
        response["nodes"] = standbys.data
    
    cs = ClusterSizing.objects.filter(db_type = DBType.MONGO.value).first()
    size_scheme = {
            Node_Size_Allowed.SMALL.name: {
                'cpu': cs.small_cpu,
                'disk': cs.small_disk,
                'memory': cs.small_memory,
                'backup': cs.small_backup_keep_days,
                'export': cs.small_export_keep_days,
            },
            Node_Size_Allowed.MEDIUM.name: {
                'cpu': cs.medium_cpu,
                'disk': cs.medium_disk,
                'memory': cs.medium_memory,
                'backup': cs.medium_backup_keep_days,
                'export': cs.medium_export_keep_days,
            },
            Node_Size_Allowed.LARGE.name: {
                'cpu': cs.large_cpu,
                'disk': cs.large_disk,
                'memory': cs.large_memory,
                'backup': cs.large_backup_keep_days,
                'export': cs.large_export_keep_days,
            },
            Node_Size_Allowed.VERY_LARGE.name: {
                'cpu': cs.very_large_cpu,
                'disk': cs.very_large_disk,
                'memory': cs.very_large_memory,
                'backup': cs.very_large_backup_keep_days,
                'export': cs.very_large_export_keep_days,
            }
        }
    response["avail_schemes"] = size_scheme
    response["databases"] = c123_DatabaseSerializer(c123_database.objects.filter(node = node, ind_deleted = False), many = True).data

    client = create_mongo_connection(node)
    if isinstance(client, tuple):
        response["ind_error"] = True
        response["detail"] = str(client[1])
        response['node'] = c123_NodeSerializer(c123_node.objects.filter(pk = id,  type = DBType.MONGO.value).first()).data
        response['node']['ex_params']['avail_change'] =  get_avail_change_date(node)
        return JsonResponse(response, status = 401)
    else:
        response["ind_error"] = c123_node.objects.filter(type = DBType.MONGO.value, cluster = node.cluster, ind_deleted = False, severity__in=[Severity.WARNING.value, Severity.NOT_OK.value]).exists() or c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).count() < 3
        response['node']['ex_params']['avail_change'] =  get_avail_change_date(node)
    try:
        #get identifier of currently connected database
        response["db_id"] = client.get_database().name
    except Exception as e:
        error = str(e)
        if "no default" in error.lower():
            response["db_id"] = "admin"
        else:
            response["detail"] = error
            return JsonResponse(response, status = 403)
    try:
    #get identifier of currently connected database
        version = client.admin.command({'buildInfo':1})['version']
        node.version = version.split('.')[0] + '.' + version.split('.')[1]
        node.save()
        response["version"] = version
    except Exception as e:
        error = str(e)
        response["detail"] = error
        return JsonResponse(response, status = 403)
    try:
        #logging verbosity level of node (1-5)
        response["log_level"] = client.admin.command({'getParameter': 1, 'logLevel': 1})["logLevel"]
    except Exception as e:
        error = str(e)
        response["detail"] = error
        return JsonResponse(response, status = 403)
    try:
        #last 10 global events as json strings
        read_only = client.admin.command({ "hello" : 1 })["readOnly"]
    except Exception as e:
        error = str(e)
        response["detail"] = error
        return JsonResponse(response, status = 403)
    
    response['read_only'] = "no" if not read_only else "yes"
    try:
        #last 10 global events as json strings
        last_50_logs = client.admin.command({ "getLog" : "global" })["log"][-50:]
    except Exception as e:
        error = str(e)
        response["detail"] = error
        return JsonResponse(response, status = 403)
    
    try:
        if float(node.version) >= 4.4:
            response["history"] = [json.loads(log, strict = False) for log in last_50_logs if log != '']
        else:
            response['history'] = []
            for log in last_50_logs:
                log = log.split(']')
                message = log[1].strip()
                stats = log[0].split(' ')
                date = stats[0]
                log_id = stats[1]
                type = stats[3]
                context = stats[5].strip('[')
                response['history'].append({
                    't': {'date': date},
                    'msg': message,
                    'id': log_id,
                    'c':type,
                    'ctx': context, 
                })
    except Exception as e:
        pass

    client = create_ssh_connection(node.host.name, node.host.username, node.host.password, node.host.private_key, node.host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            response['detail'] = client[1].strerror
        else: 
            response['detail'] = str(client[1])
        return JsonResponse(response, status = 404)
    
    jsn = json.loads(remote_command(client, f"sudo du -bs  '{node.dir_123Cluster}' | awk '{{print $1}}'", True), strict=False)
    
    if jsn['stdout'] != '':
        print(jsn['stdout'])
    
    if jsn['stderr'] != '':
        response["detail"] = jsn['stderr']
        return JsonResponse(response, status = 404)
    
    node.size = round(float(jsn['stdout'].strip()) / 1024, 2)
    node.save()
    response["display_size"] = format_size(node.size)

    jsn = json.loads(remote_command(client, f"""sudo netstat -tunlp | grep '{node.port}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}'  | sort -u | xargs -I{{}} bash -c 'awk "/^VmRSS:/ {{print \$2}}" /proc/{{}}/status && awk "{{print (\$14+\$15)/100}}" /proc/{{}}/stat' | awk '{{total_mem+=$1; total_cpu+=$2}} END {{print total_cpu, total_mem}}'""", True), strict=False)
    if jsn['stdout'] != '':
        print(jsn['stdout'])
    
    if jsn['stderr'] != '':
        response["detail"] = jsn['stderr']
        return JsonResponse(response, status = 404)
    
    measurements = jsn['stdout'].split(' ')
    response["cpu_percent"] = measurements[0]
    response["display_ram"] = format_size(measurements[1])

    return JsonResponse(response, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def AddMongoDB(request):
    current_user = request.user 
    db_name = request.data.get('db_name')
    if not db_name or db_name.isspace():
          raise APIException406("database name cannot be blank")
    db_name = ''.join(db_name.split())
    node_id = request.data.get('node_id')
    api_call = request.data.get('rest_api')
    if api_call:
        node = c123_node.objects.filter(name = node_id, type = DBType.MONGO.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type = DBType.MONGO.value, ind_deleted = False).first()
    if not node:
        raise APIException410("node id {} not available".format(node_id))
    allow_or_403(node)
    host = node.host
    auto_delete = request.data.get('auto_delete')
    auto_delete_days = request.data.get('auto_delete_days')
    
    existing_dbs = c123_database.objects.filter(name=db_name, node=node)
    if existing_dbs.exists():
        undeleted = existing_dbs.filter(ind_deleted = False)
        if undeleted.exists():
            db = undeleted.first()
        else:
            db = existing_dbs.first()
            db.ind_deleted = False
        
        db.created_by = current_user
        db.last_update_by = current_user
        db.status = Status.CREATING.value
        db.database_user = node.database_user
        db.database_pwd = node.database_pwd
        db.type = DBType.MONGO.value
        db.save()
        
    else:    
        db = c123_database.objects.create(name=db_name,
            type=DBType.MONGO.value,
            created_by=current_user,
            last_update_by=current_user,
            status=Status.CREATING.value,
            severity=Severity.NOT_OK.value,
            database_user=node.database_user,
            database_pwd=node.database_pwd,
            node = node)
    tmp_data = json.dumps({'mongocluster_data': object_to_json(node), 'host_data': object_to_json(host), 'db': object_to_json(db), 'auto_delete': auto_delete, 'created_by': current_user.id})
    task = Task.objects.create(task='create_mongo_database_task',
                                    displayment='creating mongo db {}'.format(db_name),
                                    target_type=DBType.MONGO.value,
                                    target_id=db.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropMongoDatabase(request):
    db_id = request.data.get("db_id")
    node_id = request.data.get("node_id")
    api_call = request.data.get("rest_api")
    if api_call:
        db = c123_database.objects.filter(name = db_id, node__name = node_id, type = DBType.MONGO.value, ind_deleted = False).first()
    else: 
        db = c123_database.objects.filter(pk = db_id, type = DBType.MONGO.value, ind_deleted = False).first()
   
    if not db:
        raise APIException410("database id {} not available".format(db_id))
    
    node = db.node
    allow_or_403(node)
    db.status = Status.DROPPING.value
    db.save()
    cleanup_logs(db.pk, db.type)
    host_data = c123_HostSerializer(node.host).data
    node_data = object_to_json(node)
    db_data = {'id': db.id, 'name': db.name}
    created_by = request.user
    tmp_data = {'node_data': node_data, 'db_data': db_data, 'host_data': host_data, 'created_by': created_by.id}
    task = Task.objects.create(task='drop_mongo_database_task',
                            displayment='drop database {}'.format(db.name),
                            target_type=DBType.MONGO.value,
                            target_id=db.id,
                            org_id = node.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
         
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckMongoDatabaseName(request):
    
    primary_id = request.data.get('node_id')
    db_name = request.data.get('name')
    if not db_name or db_name.isspace():
          raise APIException406("database name cannot be blank")
    db_name = ''.join(db_name.split())
    primary = c123_node.objects.filter(pk=primary_id, type = DBType.MONGO.value, ind_deleted = False).first()
    if not primary:
        raise APIException410(f"primary id {primary_id} not available")
    if db_name in ['admin', 'local', 'config']:
        return JsonResponse({'response': f'name "{db_name}" is restricted'}, status=200)
    if(re.match(r'^[a-zA-Z0-9\.\-_]{1,64}$', db_name)):
        # try to connect to the source
        client = create_mongo_connection(primary)
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                raise APIException401(client[1].strerror)
            raise APIException401(client[1])
        try:
            db_names = client.list_database_names()
        except Exception as e:
            error = e.args[0]
            print(error)
            raise APIException403(error)
        if db_name in db_names:
             return JsonResponse({'response': "already exists"}, status=200)
        return JsonResponse({'response': "name available"}, status=200)
    else:
        return JsonResponse({'error': "invalid name"}, status=406)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def StopMongoNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get('rest_api')
    current_user = request.user
    if api_call:
        node = c123_node.objects.filter(name = node_id, type=DBType.MONGO.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type=DBType.MONGO.value, ind_deleted = False).first()
    if not node:
        raise APIException410("node id {} not available".format(node_id))
    
    primary_with_nodes = node.cluster_role == ClusterRoles.PRIMARY.value and c123_node.objects.filter(type=DBType.MONGO.value, ind_deleted = False, cluster = node.cluster).exclude(cluster_role = ClusterRoles.PRIMARY.value).exists()
    if primary_with_nodes:
        secondaries = c123_node.objects.filter(type=DBType.MONGO.value, ind_deleted = False, cluster_role = ClusterRoles.STANDBY.value, cluster = node.cluster)
        secondary_data = [{"host": c123_HostSerializer(secondary.host).data, "node": object_to_json(secondary)} for secondary in secondaries]
    node.ind_started = False
    node.status = Status.STOPPING.value
    node.save()
    tmp_data = {'node_data': object_to_json(node), 'host_data': object_to_json(node.host), 'created_by': current_user.id}
    if primary_with_nodes:
        tmp_data["secondary_data"] = secondary_data
    
    task = Task.objects.create(task='stop_mongo_node_task',
                                    displayment='stopping node {}'.format(node.name),
                                    target_type=DBType.MONGO.value,
                                    target_id=node.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def StartMongoNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get('rest_api')
    current_user = request.user
    if api_call:
        node = c123_node.objects.filter(type=DBType.MONGO.value, name = node_id, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(type=DBType.MONGO.value, pk = node_id, ind_deleted = False).first()
    
    if not node:
        raise APIException410("node id {} not available".format(node_id))
    
    has_mongos = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.BALANCER.value, cluster = node.cluster, ind_deleted = False).exists()
    primary_with_nodes = node.cluster_role == ClusterRoles.PRIMARY.value and c123_node.objects.filter(type=DBType.MONGO.value, ind_deleted = False, cluster = node.cluster).exclude(cluster_role = ClusterRoles.PRIMARY.value).exists()
    if primary_with_nodes:
        secondary = c123_node.objects.filter(type=DBType.MONGO.value, ind_deleted = False, cluster_role = ClusterRoles.STANDBY.value, cluster = node.cluster)
        secondary_data = [{"id": item.pk, "name": item.name } for item in secondary]
    if node.ind_started:
         node.status = Status.RESTARTING.value
         node.save()
         tmp_data = {'node_data': object_to_json(node), 'host_data': object_to_json(node.host), 'has_mongos': has_mongos, 'created_by': current_user.id}
         if primary_with_nodes:
            tmp_data["secondary_data"] = secondary_data
         task = Task.objects.create(task='restart_mongo_node_task',
                                    displayment='restarting node {}'.format(node.name),
                                    target_type=DBType.MONGO.value,
                                    target_id=node.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    else:
        node.ind_started = True
        node.status = Status.STARTING.value
        node.save()
        tmp_data = {'node_data': object_to_json(node), 'host_data': object_to_json(node.host), 'has_mongos': has_mongos, 'created_by': current_user.id}
        if primary_with_nodes:
            tmp_data["secondary_data"] = secondary_data
        task = Task.objects.create(task='start_mongo_node_task',
                                    displayment='starting node {}'.format(node.name),
                                    target_type=DBType.MONGO.value,
                                    target_id=node.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropMongoNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get("rest_api")
    final_backup = request.data.get('final_backup')
    if api_call:
        node = c123_node.objects.filter(type=DBType.MONGO.value, name = node_id, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(type=DBType.MONGO.value, pk = node_id, ind_deleted = False).first()
    if not node:
        raise APIException410(f"node id {node_id} not available")
    
    task = drop_mongo_node(node, final_backup)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def StepDownMongoNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get("rest_api")
    if api_call:
        primary = c123_node.objects.filter(name = node_id, type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value , ind_deleted = False).first()
    else:
        primary = c123_node.objects.filter(pk = node_id, type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value , ind_deleted = False).first()

    if not primary:
        raise APIException410("primary id {} not available".format(node_id))
    in_stepdown = Task.objects.filter(task = 'stepdown_mongo_primary_task', target_id = primary.pk, status__in = [TaskState.PENDING.value, TaskState.RUNNING.value]).first()
    if in_stepdown:
        raise APIException406(f"primary {primary.name} are already in stepdown (task id {in_stepdown.pk})")
    secondaries = c123_node.objects.filter(type=DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, cluster = primary.cluster, ex_params__replicaset_name = primary.ex_params.get('replicaset_name'), ind_deleted = False)
    if secondaries:
        secondary_data = [{"id": i.id, "name": i.name} for i in secondaries]
    
    primary.status = Status.STEPDOWN.value
    primary.save()
    current_user = request.user
    has_mongos = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.BALANCER.value, cluster = primary.cluster, ind_deleted = False).exists()
    tmp_data = json.dumps({'primary_data': object_to_json(primary), 'primary_host': c123_HostSerializer(primary.host).data, 'secondary_data': secondary_data, 'has_mongos': has_mongos, 'created_by': current_user.id})
    task = Task.objects.create(task='stepdown_mongo_primary_task',
                                displayment='step down primary {}'.format(primary.name),
                                target_type=DBType.MONGO.value,
                                target_id=primary.id,
                                org_id = primary.org_id,
                                status=TaskState.PENDING.value,
                                schedule=Schedule.NOW.value,
                                parameters=tmp_data,
                                created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def MongoClusterStatus(request, id):
    primary = c123_node.objects.filter(pk = id, type=DBType.MONGO.value, ind_deleted = False, cluster_role = ClusterRoles.PRIMARY.value).first()
    if not primary:
        raise APIException410(f"primary id {id} not available")
    
    client = create_mongo_connection(primary)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])
    
    try:
        rsStatus = client.admin.command("replSetGetStatus")
    except Exception as e:
        error = str(e)
        raise APIException403(error)
    
    members = [{
                "id": i.get("_id", "none"),
                "name": i.get("name", "none"),
                "health": i.get("health", "none"),
                "stateStr": i.get("stateStr", "none"),
                "uptime": i.get("uptime", "none"),
                "pingMs": i.get("pingMs", "none")
               } for i in rsStatus["members"]
    ]
 
    status = {"set": rsStatus["set"], "members": members}
    
    return JsonResponse(status, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RestoreBackupMongoNode(request):
    """restore a mongo node from backup
    the call is made from client AJAX POST that contains the backup id"""
    bu_id = request.data.get('bu_id')
    # restore_option = 'this'/'latest'/'custom'
    restore_option = request.data.get('restore_option')
    restore_time = request.data.get('restore_time')
    autostart = request.data.get('autostart')
    api_call = request.data.get('rest_api')
    if api_call:
        backup = c123_inventory.objects.filter(name=bu_id, type = Inventory.BACKUP.value).first()
    else:
        backup = c123_inventory.objects.filter(pk=bu_id, type = Inventory.BACKUP.value).first()
        
    if not backup:
        raise APIException404("backup id {} not found".format(bu_id))
    node = backup.node
    if node.cluster_role == ClusterRoles.ARBITER.value:
        raise APIException403("backup is not allowed for arbiter")
    allow_or_403(node)
    primary_to_check = c123_node.objects.filter(cluster = node.cluster, type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary_to_check:
        raise APIException404(f"primary of node {node.name} not found")
    
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    
    node = primary_to_check
    master = find_replicaset_primary(primary_to_check)
    if master != primary_to_check.name:
        nodes = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster = primary_to_check.cluster).exclude(pk = primary_to_check.pk)
        for cluster_node in nodes:
            master = find_replicaset_primary(cluster_node)
            if master != 'none':
                primary = c123_node.objects.filter(name = master, type = DBType.MONGO.value, ind_deleted = False).first()
                if primary:
                    primary.cluster_role = ClusterRoles.PRIMARY.value
                    primary.ex_params["mongo_master_id"] = primary.id
                    primary.save()
                    others = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster = node.cluster)
                    
                    for other in others:
                        if other.cluster_role == ClusterRoles.PRIMARY.value:
                            other.cluster_role = ClusterRoles.STANDBY.value
                        other.ex_params["mongo_master_id"] = primary.id
                        other.save()
                    node = primary
                    break
    
    forbid_restore = Task.objects.filter(target_id = node.id, task__in = ['mongo_backup_restore_task', 'physical_backup_mongo_node_task'], status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
    if forbid_restore:
        raise APIException409(f"Already runs backup/restore for {node.name}")
    
    standbys = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = node.cluster)
    arbiters = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.ARBITER.value, ind_deleted = False, cluster = node.cluster)
    
    standbys_list = [{"host": c123_HostSerializer(i.host).data, "node": object_to_json(i)} for i in standbys]
    arbiters_list = [{"host": c123_HostSerializer(i.host).data, "node": object_to_json(i)} for i in arbiters]
   
    node.status = Status.RESTORING.value
    node.save()
    
    if RestoreOptions.THIS.value in restore_option:
        date = backup.created_date.strftime("%Y-%m-%d %H:%M:%S")
    elif RestoreOptions.CUSTOM.value in restore_option and restore_time:
        # example date = '2021-12-30 09:08:33'
        # convert date(str) to datetime and then to epoch(str)
        date = restore_time
    else:
        date = ''
    # serialize objects for the tasks
    backup_host = c123_HostSerializer(app_host).data
    backup = object_to_json(backup)
    node_data = object_to_json(node)
    host_data = c123_HostSerializer(node.host).data
    created_by = request.user
    if node.cluster_role == ClusterRoles.PRIMARY.value and c123_node.objects.filter(type = DBType.MONGO.value, cluster = node.cluster, ind_deleted = False, severity = Severity.OK.value, cluster_role = ClusterRoles.STANDBY.value).exists():
        autostart = True
            
    tmp_data = json.dumps({'primary_data': node_data, 'host_data': host_data, 'backup': backup, 'backup_host': backup_host, "arbiters_list": arbiters_list, "standbys_list": standbys_list, 'autostart': autostart, 'restore_option': restore_option, 'date': date, 'created_by': created_by.id})
    task = Task.objects.create(task='mongo_backup_restore_task',
                            displayment='restore backup mongo {}'.format(node.name),
                            target_type=DBType.MONGO.value,
                            target_id=node.pk,
                            org_id = node.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RestoreDumpMongoNode(request):
    """restore mongo node from dump backup
    the call is made from client AJAX POST that contains the dump id"""
    dump_id = request.data.get('dump_id')
    api_call = request.data.get('rest_api')
    if api_call:
        dump = c123_inventory.objects.filter(name=dump_id, type = Inventory.EXPORT.value).first()
    else:
        dump = c123_inventory.objects.filter(pk=dump_id, type = Inventory.EXPORT.value).first()
    if not dump:
        raise APIException404(f"dump id {dump_id} not found")
    node = dump.node
    if node.cluster_role == ClusterRoles.ARBITER.value:
        raise APIException403("dump is not allowed for arbiter")
    path = dump.ex_params.get('backup_path')
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    path_exists = checkPathExists(app_host, path)
    if not path_exists:
        raise APIException404(f"{path}: no such file or directory")
    node = dump.node
    primary = c123_node.objects.filter(cluster = node.cluster, type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
        raise APIException404("primary not found")
    
    is_master = find_replicaset_primary(primary) == primary.name
    if not is_master:
        nodes = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster = primary.cluster).exclude(pk = primary.pk)
        
        for cluster_node in nodes:
            master = find_replicaset_primary(cluster_node)
            
            if master != 'none':
                primary = c123_node.objects.filter(type = DBType.MONGO.value, cluster = primary.cluster, name = master,  ind_deleted = False).first()
                if not primary:
                    raise APIException410("primary not in app db")
                primary.cluster_role = ClusterRoles.PRIMARY.value
                primary.ex_params["mongo_master_id"] = primary.pk
                primary.save()
                others = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster = node.cluster, ex_params__replicaset_name = primary.ex_params.get('replicaset_name'))
                    
                for other in others:
                    if other.cluster_role == ClusterRoles.PRIMARY.value:
                        other.cluster_role = ClusterRoles.STANDBY.value
                    other.ex_params["mongo_master_id"] = primary.pk
                    other.save()
                break
    
    primary.status = Status.RESTORING.value
    primary.save()
    # serialize objects for the tasks
    dump_host = c123_HostSerializer(app_host).data
    dump = object_to_json(dump)
    node_data = object_to_json(primary)
    host_data = c123_HostSerializer(primary.host).data
    created_by = request.user
    tmp_data = json.dumps({'db_name': None, 'node_data': node_data, 'host_data': host_data, 'dump': dump, 'backup_host': dump_host, 'created_by': created_by.id})
    task = Task.objects.create(task='mongorestore_task',
                            displayment=f'restore mongo cluster {primary.name}',
                            target_type=DBType.MONGO.value,
                            target_id=primary.id,
                            org_id = primary.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckMongoDatabaseSize(request):
    node_id = request.data.get('node_id')
    db_name = request.data.get("db_name")
    node = c123_node.objects.filter(pk = node_id, type = DBType.MONGO.value, ind_deleted = False).first()
    if not node:
        raise APIException404("node id {} not found".format(node_id))
    client = create_ssh_connection(node.host.name, node.host.username, node.host.password, node.host.private_key, node.host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])
    
    mongo_client = create_mongo_connection(node)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
                    raise APIException401(client[1].strerror)
        raise APIException401(str(client[1]))
    
    else:

        try:
                dbs = mongo_client.list_databases()
        except Exception as e:
            print(e.args[0])
            raise APIException403(e.args[0])
        if db_name:
            db = next((i for i in dbs if i["name"] == db_name), None)
            if not db:
                raise APIException404(f"database {db_name} not found")
            db_size = float(db['sizeOnDisk']) / 1024
        else:
            db_size = 0
            for db in dbs:
                db_size += db['sizeOnDisk']
            db_size /= 1024
        
        
        command = 'df -P . | tail -1 | awk \'{print $4}\''
        jsn = json.loads(remote_command(client, command, True), strict=False)
        if(not jsn['stderr'] == ''):
            print(jsn['stderr'])
            return JsonResponse({'detail': jsn['stderr']}, status=404)
        
        free_space = float(jsn['stdout'])
        return JsonResponse({"size": db_size, "display_size": format_size(db_size), 'free_space': free_space}, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RemoveMongoSchedule(request):
    node_id = request.data.get("node_id")
    node = c123_node.objects.filter(pk=node_id, type = DBType.MONGO.value, ind_deleted = False).first()
    if not node:
        raise APIException410(f"primary id {node_id} not available")
    allow_or_403(node)
    backup = c123_inventory.objects.filter(node = node, type = Inventory.BACKUP.value, is_sched = True).last()
    from django.conf import settings
    if backup:
        Task.objects.filter(pk = backup.ex_params.get("oplog_task_id")).update(status = TaskState.ON_DEMAND.value)
        job_id = backup.ex_params.get("job_id")
        job = settings.scheduler.get_job(job_id)
        if job:
            settings.scheduler.remove_job(job_id)
            notification = f"{job_id} removed"
            job = settings.scheduler.get_job(f'mongodump_oplog_task_{node.cluster.pk}')
            if job:
                settings.scheduler.remove_job(f'mongodump_oplog_task_{node.cluster.pk}')
                
            return JsonResponse({"notification": notification}, status=200)
        
    else:
        standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, type = DBType.MONGO.value, cluster = node.cluster)
        if standbys.exists():
            for standby in standbys:
                backup = c123_inventory.objects.filter(node = standby, type = Inventory.BACKUP.value, is_sched = True).last()
                if backup:
                    Task.objects.filter(pk = backup.ex_params.get("oplog_task_id")).update(status = TaskState.ON_DEMAND.value)
                    job_id = backup.ex_params.get("job_id")
                    job = settings.scheduler.get_job(job_id=job_id)
                    if job:
                        settings.scheduler.remove_job(job_id=job_id)
                        notification = f"{job_id} removed"
                        job = settings.scheduler.get_job(f'mongodump_oplog_task_{node.cluster.pk}')
                        if job:
                            settings.scheduler.remove_job(f'mongodump_oplog_task_{node.cluster.pk}')
                        return JsonResponse({"notification": notification}, status=200)            
    return JsonResponse({"notification": "no schedule backup found"}, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def UpgradeMongoNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get("rest_api")
    target_version = request.data.get('target_version')
    current_user = request.user
    if api_call:
        node = c123_node.objects.filter(name = node_id, type = DBType.MONGO.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type = DBType.MONGO.value, ind_deleted = False).first()
    if not node:
        raise APIException410("node id {} not available".format(node_id))
    
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    
    nodes = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster = node.cluster)
    node_list = [{'node': object_to_json(n), 'host': object_to_json(n.host)} for n in nodes]
    
    nodes.update(status = Status.UPGRADING.value)
    
    tmp_data = {'app_host': object_to_json(app_host), 'nodes': node_list, 'target_version': target_version, 'created_by': current_user.id}
    task = Task.objects.create(task='upgrade_mongo_node_task',
                                displayment='upgrading {} to mongodb {}'.format(node.name, target_version),
                                target_type=DBType.MONGO.value,
                                target_id=node.id,
                                org_id=node.org_id,
                                status=TaskState.PENDING.value,
                                schedule=Schedule.NOW.value,
                                parameters=tmp_data,
                                created_by=current_user)
    return append_status_cmd(request, task)


@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropMongoShardedCluster(request):
    cluster_id = request.data.get("cluster_id")
    api_call = request.data.get("rest_api")
    if api_call:
        cluster = c123_cluster.objects.filter(name=cluster_id, type = DBType.MONGO.value).first()
    else:
        cluster = c123_cluster.objects.filter(pk=cluster_id, type = DBType.MONGO.value).first()
    if not cluster:
        raise APIException404(f"cluster ID {cluster_id} could not be found")

    cluster_nodes = c123_node.objects.filter(type = DBType.MONGO.value, cluster = cluster)
    non_deleted_nodes = cluster_nodes.filter(ind_deleted = False)
    if non_deleted_nodes.exists():
        if non_deleted_nodes.exclude(cluster_role = ClusterRoles.BALANCER.value).exists():
            raise APIException409(f"non-standalone cluster {cluster.name} has nodes in use.Could not be dropped")
        else:
            mongos_node = non_deleted_nodes.first()
            task = drop_mongo_node(mongos_node)
            return append_status_cmd(request, task)
    else:
        
        for node in cluster_nodes:
            node.cluster = None
            node.save()
        
        cluster_name = cluster.name
        cluster.delete()
    
        return JsonResponse({"notification": f"cluster {cluster_name} has been dropped"}, status=200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def BackupMongoNode(request):
    """backup for mongo cluster"""
    schedule = request.data.get('schedule')
    node_id = request.data.get('node_id')
    compress = request.data.get('compress')
    api_call = request.data.get('rest_api')
    from django.conf import settings
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    if api_call:
        mongo_node = c123_node.objects.filter(type = DBType.MONGO.value, name=node_id, ind_deleted = False).first()
    else:
        mongo_node = c123_node.objects.filter(type = DBType.MONGO.value, pk=node_id, ind_deleted = False).first()
    if not mongo_node:
        raise APIException406(f"node id {node_id} not available for backup")
    allow_or_403(mongo_node)
    primary_with_nodes = mongo_node.cluster_role == ClusterRoles.PRIMARY.value and c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster = mongo_node.cluster).exclude(cluster_role = ClusterRoles.PRIMARY.value).exists()
    if primary_with_nodes:
        secondary = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster_role = ClusterRoles.STANDBY.value, cluster = mongo_node.cluster)
        arbiters = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False, cluster_role = ClusterRoles.ARBITER.value, cluster = mongo_node.cluster)
        arbiters_id = [item.pk for item in arbiters]
        secondary_data = [{"id": item.pk, "name": item.name } for item in secondary]
    else:
        secondary_data = []
    backup_path = request.data.get('backup_path')
    update_schedule = request.data.get("update_schedule")
    if schedule:
        client_timezone = pytz.timezone(request.data.get("client_timezone"))
        schedule_run_time_str = request.data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = request.data.get("days_of_week")
        days_of_month = request.data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
    
    primary = c123_node.objects.filter(cluster = mongo_node.cluster, cluster_role = ClusterRoles.PRIMARY.value, type = DBType.MONGO.value, ind_deleted = False).first()
    if not primary:
        raise APIException410(f"primary of {mongo_node.name} not available")
    if update_schedule and schedule:
        backup = c123_inventory.objects.filter(node = primary, type = Inventory.BACKUP.value, is_sched = True).last()
        if backup:
            backup.ex_params['description'] = description
            backup.ex_params['schedule_run_time'] = schedule_run_time_str
            backup.ex_params['days_of_month'] = days_of_month
            backup.ex_params['days_of_week'] = days_of_week
            backup.save()
            notification = f"{backup.node} backup updated to '{description}'"
            job_id = backup.ex_params.get("job_id")
            job = settings.scheduler.get_job(job_id=job_id)
            if job:
                job.reschedule(trigger=trigger)
            else:
                settings.scheduler.add_job(schedule_backup, trigger=trigger, kwargs={"task_id":  backup.ex_params["task_id"]}, replace_existing=True, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP, id=job_id)
                settings.scheduler.add_job(mongodump_oplog_job, IntervalTrigger(minutes=1, timezone=settings.TIME_ZONE), replace_existing=True, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP, id = f'mongodump_oplog_task_{primary.cluster.pk}',  kwargs={"cluster_id": primary.cluster.pk})
            Task.objects.filter(pk = backup.ex_params.get("oplog_task_id")).update(status = TaskState.PENDING.value)
            
            return JsonResponse({"notification": notification}, status=200)
        
        else:
            standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.MONGO.value, ind_deleted = False, cluster = primary.cluster)
            if standbys.exists():
                for standby in standbys:
                    backup = c123_inventory.objects.filter(node = standby.pk, type = Inventory.BACKUP.value, is_sched = True).last()
                    if backup:
                        backup.ex_params['schedule_run_time'] = schedule_run_time_str
                        backup.ex_params['days_of_month'] = days_of_month
                        backup.ex_params['days_of_week'] = days_of_week
                        backup.ex_params['description'] = description
                        backup.save()
                        notification = f"{backup.node} backup updated to '{description}'"
                        job_id = backup.ex_params.get("job_id")
                        job = settings.scheduler.get_job(job_id=job_id)
                        if job:
                           job.reschedule(trigger=trigger)
                        else:
                            settings.scheduler.add_job(schedule_backup, trigger=trigger, kwargs={"task_id":  backup.ex_params["task_id"]}, replace_existing=True, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP,  id=job_id)
                            settings.scheduler.add_job(mongodump_oplog_job, IntervalTrigger(minutes=1, timezone=settings.TIME_ZONE), replace_existing=True, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP, id = f'mongodump_oplog_task_{primary.cluster.pk}',  kwargs={"cluster_id": primary.cluster.pk})
                        Task.objects.filter(pk = backup.ex_params.get("oplog_task_id")).update(status = TaskState.PENDING.value)
                        return JsonResponse({"notification": notification}, status=200)
                
                return JsonResponse({"notification": "no schedule backup found"}, status = 200)            
            else:
               return JsonResponse({"notification": "no schedule backup found"}, status = 200)
    serializer = c123_inventory_serializer(data = request.data, partial = True)
    if serializer.is_valid():
        if mongo_node.cluster_role == ClusterRoles.ARBITER.value:
           raise APIException403("backup is not allowed for arbiter node")
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MONGO.value)
        keep_days = getattr(size_scheme, f'{mongo_node.sizing.lower()}_backup_keep_days', None)
        check_space(node = mongo_node, path=backup_path, source_path=mongo_node.dir_123Cluster)
    # create the task signature
        
        if schedule:
            existing_backup = c123_inventory.objects.filter(node = mongo_node, type = Inventory.BACKUP.value, is_sched = True)
            if existing_backup.exists():
                raise APIException409("Schedule backup already exists.\nCan be only one schedule backup for cluster.")
            
            if mongo_node.cluster_role == ClusterRoles.PRIMARY.value:
                standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.MONGO.value, ind_deleted = False, cluster = mongo_node.cluster)
            else:
                standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.MONGO.value, ind_deleted = False, cluster = mongo_node.cluster).exclude(pk = mongo_node.pk)
            
            if standbys.exists():
                for standby in standbys:
                    existing_backup = c123_inventory.objects.filter(node = standby, type = Inventory.BACKUP.value, is_sched = True)
                    if existing_backup.exists():
                        raise APIException409("Schedule backup already exists.\nCan be only one schedule backup for cluster.")
            backup = serializer.create(serializer.validated_data)
            backup.node = mongo_node
            if keep_days:
                backup.keep_days = keep_days
            backup.type = Inventory.BACKUP.value
            backup.created_by = request.user
            backup.ex_params = {"size": 0}
            backup.ex_params["belong_to"] = DBType.MONGO.value
            backup.ex_params["node_version"] = mongo_node.version
            backup.is_sched = schedule
            backup.ex_params["duration"] = 0
            backup.ex_params["compress"] = compress
            backup.ex_params["modified_date"] = ''
            backup.ex_params['description'] = description
            backup.ex_params['schedule_run_time'] = schedule_run_time_str
            backup.ex_params['days_of_month'] = days_of_month
            backup.ex_params['days_of_week'] = days_of_week
            backup.ex_params["oplog_path"] = ''
            backup.ex_params["oplog_time"] = ''
            notification =  "{} backup {}".format(backup.node, description)
            job_id = f"{notification} ({client_timezone})"
            backup.ex_params["job_id"] = job_id
            backup.ex_params["backup_path"] = f'{backup_path}/{mongo_node.host.name}_{mongo_node.port}'
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            backup.name = "{}_{}".format(mongo_node.name, now)
            backup.save()
           
            # serialize objects for the tasks
            node_data = object_to_json(mongo_node)
            host_data = c123_HostSerializer(mongo_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            created_by = request.user
            tmp_data = {'node_data': node_data, 'host_data': host_data, 'backup_host':backup_host, 'backup_data': object_to_json(backup), 'created_by': created_by.id}
            if primary_with_nodes:
                tmp_data["arbiters_id"] = arbiters_id
                tmp_data["secondary_data"] = secondary_data
                standbys = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = mongo_node.cluster)
                arbiters = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = mongo_node.cluster)
                
                standbys_list = [{"host": c123_HostSerializer(i.host).data, "node": object_to_json(i)} for i in standbys]
                arbiters_list = [{"host": c123_HostSerializer(i.host).data, "node": object_to_json(i)} for i in arbiters]
                
                tmp_data["standbys"] = standbys_list
                tmp_data["arbiters"] = arbiters_list
            
            backup.save()
             
            task = Task.objects.create(task='physical_backup_mongo_node_task',
                                            displayment='backup {}'.format(mongo_node.name), 
                                            target_type=DBType.MONGO.value, 
                                            target_id=mongo_node.id,
                                            org_id = mongo_node.org_id,
                                            schedule=Schedule.SCHEDULED.value, 
                                            status=TaskState.ON_DEMAND.value,
                                            parameters=json.dumps(tmp_data),
                                            created_by=created_by)
            
            job = settings.scheduler.add_job(schedule_backup, trigger=trigger, kwargs={"task_id": task.pk}, replace_existing=True, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP, id=job_id)

            primary_data = object_to_json(primary)
            host_primary = c123_HostSerializer(primary.host).data
            tmp_data = json.dumps({'primary_data': primary_data, 'host_primary': host_primary, 'backup_host': backup_host, 'backup_data': object_to_json(backup), 'created_by': created_by.id})
            task2 = Task.objects.create(task=f'mongodump_oplog_task',
                                            displayment='mongodump oplog {}'.format(primary.name), 
                                            target_type=DBType.MONGO.value, 
                                            target_id=primary.id,
                                            org_id = primary.org_id, 
                                            schedule=Schedule.MINUTES_15.value, 
                                            status=TaskState.PENDING.value,
                                            parameters=tmp_data,
                                    created_by=created_by)
             
            oplog_job_id = f'{task2.task}_{primary.cluster.pk}'
            settings.scheduler.add_job(mongodump_oplog_job, IntervalTrigger(minutes=1, timezone=settings.TIME_ZONE), replace_existing=True, next_run_time=timezone.now() + timedelta(seconds=300), misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP, id = oplog_job_id,  kwargs={"cluster_id": primary.cluster.pk})
            backup.ex_params['task_id'] = task.pk
            backup.ex_params["oplog_task_id"] = task2.pk
            backup.ex_params["oplog_job_id"] = oplog_job_id
            backup.save()               
        else:
            existing_backup = Task.objects.filter(target_id = mongo_node.id, task = 'physical_backup_mongo_node_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
            if existing_backup:
                raise APIException409(f"Already runs backup for {mongo_node.name}")
            backup = serializer.create(serializer.validated_data)
            backup.node = mongo_node
            if keep_days:
                backup.keep_days = keep_days
            backup.type = Inventory.BACKUP.value
            backup.created_by = request.user
            backup.ex_params = {"size": 0}
            backup.ex_params["belong_to"] = DBType.MONGO.value
            backup.ex_params["node_version"] = mongo_node.version
            backup.is_sched = schedule
            backup.ex_params["duration"] = 0
            backup.ex_params["compress"] = compress
            backup.ex_params["modified_date"] = ''
            backup.ex_params["oplog_path"] = ''
            backup.ex_params["oplog_time"] = ''
            # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            backup.name = "{}_{}".format(mongo_node.name, now)
            backup.ex_params["backup_path"] = f'{backup_path}/physical/{mongo_node.host.name}_{mongo_node.port}/{now}'
            backup.save()
            notification = f"backup of {mongo_node.name}"
            # serialize objects for the tasks
            node_data = object_to_json(mongo_node)
            host_data = c123_HostSerializer(mongo_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            created_by = request.user
            tmp_data = {'node_data': node_data, 'host_data': host_data, 'backup_host': backup_host, 'backup_data': object_to_json(backup), 'compress': compress, 'created_by': created_by.id}
            if primary_with_nodes:
                tmp_data["arbiters_id"] = arbiters_id
                tmp_data["secondary_data"] = secondary_data
                standbys = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = mongo_node.cluster)
                arbiters = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = mongo_node.cluster)
                standbys_list = [{"host": c123_HostSerializer(i.host).data, "node": object_to_json(i)} for i in standbys]
                arbiters_list = [{"host": c123_HostSerializer(i.host).data, "node": object_to_json(i)} for i in arbiters]
                tmp_data["standbys"] = standbys_list
                tmp_data["arbiters"] = arbiters_list
            
            task = Task.objects.create(task='physical_backup_mongo_node_task',
                                            displayment='backup mongo node {}'.format(mongo_node.name), 
                                            target_type=DBType.MONGO.value, 
                                            target_id=mongo_node.id,
                                            org_id = mongo_node.org_id, 
                                            schedule=Schedule.NOW.value, 
                                            status=TaskState.PENDING.value,
                                            parameters=json.dumps(tmp_data),
                                            created_by=created_by)
        
        backup.ex_params['task_id'] = task.pk
        backup.save()
    else:
        print("invalid form")
        print(serializer.errors)        
        return JsonResponse({'error': serializer.errors}, status=404)
    return append_status_cmd(request, task, notification)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DumpMongoNode(request):
    """dump for mongo node"""
    data = request.data
    data['created_by'] = request.user.id
    schedule = data.get('schedule')
    api_call = data.get('api_call')
    node_id = data.get('node_id')

    if not schedule:
        task = export_mongo_node(data)
        return append_status_cmd(request, task)
    else:
        from django.conf import settings
        client_timezone = pytz.timezone(data.get("client_timezone"))
        schedule_run_time_str = data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = data.get("days_of_week")
        days_of_month = data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
        if api_call:
            node = c123_node.objects.filter(type = DBType.MONGO.value, name=node_id, ind_deleted = False).first()
        else:
            node = c123_node.objects.filter(type = DBType.MONGO.value, pk=node_id, ind_deleted = False).first()
        if not node:
            raise APIException410("node id {} not available".format(node_id))
        job_id =  f"export_{node.name}"
        scheduler = settings.scheduler
        export_job = next((i for i in scheduler.get_jobs() if job_id in i.id), None)
        if export_job:
            export_job.reschedule(trigger=trigger)
        else:
             settings.scheduler.add_job(export_mongo_node, replace_existing=True, trigger=trigger,  kwargs={"data": data}, id=job_id, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        return JsonResponse({"notification": f"export {node.name} {description} started"}, status = 200)
    
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckMongoVersionExists(request):
    """check if a mongo version package already exists"""
    host_name = request.data.get("host")
    username = request.data.get("os_username")
    password = request.data.get("os_password")
    if password:
        password = format_password(password)

    private_key = request.data.get("private_key")
    ssh_port = request.data.get("ssh_port")
    version = request.data.get("software_version")
        
    host = SimpleNamespace( **{
        "name": host_name, 
        "username": username, 
        "password": password,
        "private_key": private_key,
        "ssh_port": ssh_port
    })
    client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])
    jsn = json.loads(remote_command(client, f"rpm -qa | grep mongodb-org-server-{version}", True))
    if(jsn['stderr'] != ''):
        print(jsn['stderr'])
        if 'not found' in jsn['stderr']:
            jsn = json.loads(remote_command(client, f"dpkg -l | grep mongodb-org-server | grep {version}", True))
            
            if(jsn['stderr'] != ''):                
                return JsonResponse({'error': jsn['stderr']}, status=404)
            elif(jsn['stdout'] != ''):
                return JsonResponse({'exists': True}, status=200)
    if(jsn['stdout'] != ''):
        return JsonResponse({'exists': True}, status=200)
        
    return JsonResponse({'exists': False}, status=200)
